#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stddef.h>

#include <sysutil/sysutil_sysparam.h>
#include <sys/timer.h>

#include <PSGL/psgl.h>
#include <PSGL/psglu.h>
#include <cell/dbgfont.h>

#include "cellgraphicspsglutil.h"

static bool keep_aspect = true;
static uint8_t *gl_buffer;
static GLuint texture;
static GLuint tex_filter;
static PSGLdevice* psgl_device = NULL;
static PSGLcontext* psgl_context = NULL;
static uint16_t* align_buf = NULL;
/* defines if GL_LINEAR is set, otherwise GL_NEAREST
 * should of course become far more sophisticated,
 * allowing for a whole range of values. Perhaps
 * use the straight Enum constants Sony uses
 * look at psglCreateDeviceAuto for an example
*/
static bool smooth = true;

static GLuint gl_width = 0;
static GLuint gl_height = 0;



static CellDbgFontConsoleId dbg_id;


void cellPSGLUtilInitDbgFont()
{
	CellDbgFontConfig cfg;
	memset(&cfg, 0, sizeof(cfg));
	cfg.bufSize = 512;
	cfg.screenWidth = gl_width;
	cfg.screenHeight = gl_height;
	cellDbgFontInit(&cfg);

	CellDbgFontConsoleConfig ccfg;
	memset(&ccfg, 0, sizeof(ccfg));

	ccfg.posLeft	=	0.1f;
	ccfg.posTop	=	0.1f;
	ccfg.cnsWidth	=	32;
	ccfg.cnsHeight	=	16;
	ccfg.scale	=	1.0f;
	ccfg.color	=	0xffffffff;	//ABGR -> white
	dbg_id = cellDbgFontConsoleOpen(&ccfg);
}

void cellPSGLUtilDeinitDbgFont()
{
	cellDbgFontConsoleClose(dbg_id);
	cellDbgFontExit();
}

/*
 * This is the font rendering function in console mode
 * I think this would work best for the rendering of the ROM menu
 * Perhaps experiment
 * Also - a better name for it
*/
void dprintf_console(const char* fmt, ...)
{
	cellPSGLUtilClearSurface();

	va_list ap;

	va_start(ap, fmt);
	cellDbgFontConsoleVprintf(dbg_id, fmt, ap);
	va_end(ap);

	cellDbgFontConsoleEnable(dbg_id);
	cellDbgFontDraw();
	psglSwap();	// added comment by SSNES author - should this be here?
}

/*
 * This is the font rendering function in label mode (I presume)
 * We used this for the rendering of all the debug messages in SSNES - bSNES
 * Probably not suitable for display of ROM screen
 * Also - a better name for it
*/

void gl_dprintf(float x, float y, float scale, const char* fmt, ...)
{
	cellPSGLUtilClearSurface();

	char buffer[512];

	va_list ap;

	va_start(ap, fmt);
	vsnprintf(buffer, 512, fmt, ap);
	cellDbgFontPuts(x, y, scale, 0xffffffff, buffer);
	va_end(ap);

	cellDbgFontDraw();
	psglSwap();
}

void cellPSGLUtilClearSurface()
{
	glClear(GL_COLOR_BUFFER_BIT);
}

/*
 * This is the main render function - taken straight
 * from SSNES - bSNES
 * Perhaps needs to be changed a bit - perhaps no 
 * need for all the arguments
*/
static void cellPSGLUtilRender(const uint16_t* frame, int width, int height)
{
	S9X_DBG("glClearColor(0.0f, 0.0f, 0.0f, 1.0f)");
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	S9X_DBG("cellPSGLUtilClearSurface");
	cellPSGLUtilClearSurface();

	S9X_DBG("glMatrixMode(GL_MODELVIEW)");
	glMatrixMode(GL_MODELVIEW);
	S9X_DBG("glLoadIdentity)");
	glLoadIdentity();

	//straight carbon copy from SSNES - bSNES -
	//itself pilfered from an SDK example
	//perhaps needs to be scrutinized more closely
	S9X_DBG("glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER)");
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
	S9X_DBG("glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER)");
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
	S9X_DBG("glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, tex_filter)");
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, tex_filter);
	S9X_DBG("glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, tex_filter)");
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, tex_filter);

	S9X_DBG("glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_BGRA, GL_UNSIGNED_SHORT_1_5_5_5_REV, frame)");
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_BGRA, GL_UNSIGNED_SHORT_1_5_5_5_REV, frame);

	S9X_DBG("glLoadIdentity");
	glLoadIdentity();
	S9X_DBG("glColor4f(1.0, 1.0");
	glColor4f(1.0, 1.0, 1.0, 1.0);

	// width of default SNES height (224) divided by default SNES width (256)

	S9X_DBG("vertexes[]");
	GLfloat vertexes[] = {
		0.0, 0.0, 0.0,
		0.0, 1.0, 0.0,
		1.0, 1.0, 0.0,
		1.0, 0.0, 0.0
	};

	S9X_DBG("tex_coords[]");
	GLfloat tex_coords[] = {
		0, (float)height/512,
		0, 0,
		(float)width/512, 0,
		(float)width/512, (float)height/512
	};

	S9X_DBG("glEnableClientState(GL_VERTEX_ARRAY)");
	glEnableClientState(GL_VERTEX_ARRAY);
	S9X_DBG("glEnableClientState(GL_TEXTURE_COORD_ARRAY)");
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	S9X_DBG("glVertexPointer(3, GL_FLOAT, 3 * sizeof(GLfloat), vertexes)");
	glVertexPointer(3, GL_FLOAT, 3 * sizeof(GLfloat), vertexes);
	S9X_DBG("glTexCoordPointer(2, GL_FLOAT, 2 * sizeof(GLfloat), tex_coords)");
	glTexCoordPointer(2, GL_FLOAT, 2 * sizeof(GLfloat), tex_coords);

	S9X_DBG("glDrawArrays(GL_QUADS, 0, 4)");
	glDrawArrays(GL_QUADS, 0, 4); // added comment from bSNES - SSNES - is this correct?

	S9X_DBG("glDisableClientState(GL_VERTEX_ARRAY)");
	glDisableClientState(GL_VERTEX_ARRAY);
	S9X_DBG("glDisableClientState(GL_TEXTURE_COORD_ARRAY)");
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	S9X_DBG("psglSwap()");
	psglSwap(); // Swaps the render to screen;

}

static inline void process_frame (uint16_t * __restrict__ out, const uint16_t * __restrict__ in, unsigned width, unsigned height)
{
   unsigned pitch = 1024;
   if ( height == 448 || height == 478 )
      pitch = 512;

   for ( unsigned y = 0; y < height; y++ )
   {
      const uint16_t *src = in + y * pitch;
      uint16_t *dst = out + y * width;

      memcpy(dst, src, width * sizeof(uint16_t));
   }
}

void cellPSGLUtilFrame(const uint16_t* frame, unsigned width, unsigned height)
{
   S9X_DBG("process_frame(output, frame, width, height)");
   process_frame(align_buf, frame, width, height);
   S9X_DBG("cellPSGLUtilRender(output, width, height)");
   cellPSGLUtilRender(align_buf, width, height);
}

static void cellPSGLUtilSetViewports()
{
	unsigned int width = gl_width, height = gl_height;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	if ( keep_aspect )
	{
		float desired_aspect = 296.0/224.0;
		float in_aspect = (float)width / height;

		if ( (int)(in_aspect*100) > (int)(desired_aspect*100) )
		{
			float delta = (in_aspect / desired_aspect - 1.0) / 2.0 + 0.5;
			glOrthof(0.5 - delta, 0.5 + delta, 0, 1, -1, 1);
		}

		else if ( (int)(in_aspect*100) < (int)(desired_aspect*100) )
		{
			float delta = (desired_aspect / in_aspect - 1.0) / 2.0 + 0.5;
			glOrthof(0, 1, 0.5 - delta, 0.5 + delta, -1, 1);
		}
		else
			glOrthof(0, 1, 0, 1, -1, 1);
	}
	else
		glOrthof(0, 1, 0, 1, -1, 1);

	glViewport(0, 0, width, height);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void cellPSGLUtilInitDisplay () {

	if (smooth)
		tex_filter = GL_LINEAR;
	else
		tex_filter = GL_NEAREST;

	PSGLinitOptions options;
	options.enable = PSGL_INIT_MAX_SPUS | PSGL_INIT_INITIALIZE_SPUS;
	options.maxSPUs = 1;
	options.initializeSPUs = false;

	psglInit(&options);

	psgl_device = psglCreateDeviceAuto(GL_ARGB_SCE, GL_DEPTH_COMPONENT24, GL_MULTISAMPLING_4X_SQUARE_ROTATED_SCE);

	//Perhaps remove in future

	psglGetDeviceDimensions(psgl_device, &gl_width, &gl_height);

	//Perhaps remove in future

	psgl_context = psglCreateContext();
	psglMakeCurrent(psgl_context, psgl_device);

	psglResetCurrentContext();

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DITHER);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_SCISSOR_TEST);

	//Set viewport function comes here cellPSGLUtilSetViewPort
	cellPSGLUtilSetViewports();

	//video->input_scale is scaling for use with filters - should be 4 with HQ2X and 8 with HQ4X, 8 with NTSC, and 2 for the rest
	gl_buffer   = (uint8_t*)memalign(128, 256 * 256 * 2 * 2 * 2);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	//need to set up temporary buffer for first init
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	//glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 256 * video->input_scale, 256 * video_input_scale, 0, GL_BGRA, GL_UNSIGNED_SHORT_1_5_5_5_REV, gl_buffer); 
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 256 * 2, 256 * 2, 0, GL_BGRA, GL_UNSIGNED_SHORT_1_5_5_5_REV, gl_buffer); 

	//PSGL doesn't clear the screen on startup - so this needs to be done
	cellPSGLUtilClearSurface();
	psglSwap();

   cellPSGLUtilInitDbgFont();
   align_buf = (uint16_t*)memalign(128, 512 * 512 * sizeof(uint16_t));
}

void cellPSGLUtilDeinitDisplay(void *data)
{
   free(align_buf);
   cellPSGLUtilDeinitDbgFont();
	glDeleteTextures(1, &texture);
	free(gl_buffer);

	if (psgl_context)
	{
		psglDestroyContext(psgl_context);
		psgl_context = NULL;
	}
	if (psgl_device)
	{
		psglDestroyDevice(psgl_device);
		psgl_device = NULL;
	}
	psglExit();
}
