#ifndef CELLGRAPHICSUTIL_H_
#define CELLGRAPHICSUTIL_H_

#include <stdint.h>
#include <stdlib.h>
#include <stdbool.h>
#include <cell/gcm.h>
#include <cell/dbgfont.h>

#include <sys/cdefs.h>
#include <sysutil/sysutil_sysparam.h>

#define CMD_BUFFER_SIZE (1024 * 64)
#define IO_BUFFER_SIZE (1024 * 1024)
#define COLOR_BUFFER_NUM 2
#define SHADER(x) extern uint32_t _binary_##x##_start; static unsigned char *shader_##x##_ptr = (unsigned char *)&_binary_##x##_start; CGprogram shader_##x; static void *shader_##x##_ucode; static uint32_t shader_##x##_offset

// define some colours for dgbfont output (remember it uses ABGR colour format)
#define WHITE	0xffffffff
#define RED	0xff0000ff
#define GREEN	0xff00ff00
#define BLUE	0xffff0000
#define YELLOW	0xff00ffff
#define PURPLE	0xffff00ff
#define CYAN	0xffffff00

#define FONT_SIZE 1.0f
#define DEBUG_FONT_SIZE 1.0f

#define FONT_HEIGHT 9
#define FONT_WIDTH 8
#define FONT_SCALE 1

#ifdef S9XDEBUG
#define S9X_DBG(fmt, args...) do {\
	cellDbgFontPrintf(0.1f, 0.1f, DEBUG_FONT_SIZE, RED, fmt, ##args);\
   sys_timer_usleep(2000000);\
   } while(0);
#else
#define S9X_DBG(fmt, args...) ((void)0)
#endif


typedef struct {
	void *address;
	uint32_t offset;
	uint32_t pitch;
	uint32_t width;
	uint32_t height;
} color_buf_t;

typedef struct {
	uint8_t a;
	uint8_t r;
	uint8_t g;
	uint8_t b;
} color4_t;

extern color4_t *display;
extern color4_t *display_back;
extern color_buf_t color_buf[COLOR_BUFFER_NUM];

extern void cellGcmUtilWaitFlip();

extern void cellGcmUtilFlip();

extern void cellGcmUtilInitShader();

extern void cellGcmUtilSetDrawEnv(void);

extern void cellGcmUtilSetRenderColor();

extern void cellGcmUtilRender(bool sync);

extern void cellGcmUtilSetRenderTarget(const uint32_t Index);

extern void cellGcmUtilClearSurface();

extern int32_t cellGcmUtilInitDisplay();

extern int32_t cellGcmUtilInitDbgFont();

extern int32_t cellChangeResolution(uint32_t resId);
extern CellVideoOutState CellGetVideoState();

#endif
